c 
c  Copyright 1999, 2001 Illinois Institute of Technology
c 
c  See the file "LICENSE" for information on usage and redistribution
c  of this file, and for a DISCLAIMER OF ALL WARRANTIES.
c 
c
c  This program fits a step function to the data from an optimization scan.
c  In other words,
c
c  f(x) = a1  ( for x <  a3 )
c  f(x) = a2  ( for x => a3 )
c
c  The LMDIF1 routine works in terms of the array index i rather than
c  the position variable x, so we must redefine x as a function of j by
c
c  x(j) = x0 + x1 * j
c
c  Substituting this for x in the first equation, we get the new expression
c
c  f(x) = b(1) ( for x <  b(3) )
c  f(x) = b(2) ( for x => b(3) )
c
c  where
c
c  b(1) = a1 
c  b(2) = a2
c  b(3) = (a3 - x0) / x1
c
c  In all of this, j is a zero based array index.  Since Fortran uses 1-based
c  array indices, the two are related by j = i - 1.
c
c  The results we ultimately want are the parameters a1, a2, and a3.
c  Since the values of x0 and x1 are input parameters for the program, we
c  can invert this to get
c
c  a1 = b(1)
c  a2 = b(2)
c  a3 = x0 + x1 * b(3)
c
c  The LMDIF1 routine is described on pages 115-121 of the 
c  "User Guide for Minpack-1", ANL-80-74.
c
      program fit_to_step_function
      implicit none

      character*256 filename
      real*8 a(3), b(3)
      real*8 function_array(10000)
      real*8 x0, x1, x_simple, measured_data
      real*8 algorithm_tolerance, euclidean_norm
      real*8 xstart, xnext, xvalue(10000)
      real*8 min_position, min_value, max_position, max_value

      integer closest_to_half_max, second_closest_to_half_max
      real*8 half_max_value, half_max_position
      real*8 distance, smallest_distance, second_smallest_distance
      real*8 xa, xb, ya, yb, slope

      integer i, num_points, num_variables, errflag
      integer algorithm_info, state

      integer int_work_array(3)
      real*8 work_array(60025)
      integer length_work_array

      real*8 dpmpar, enorm
      external fit_function

      common /optimize/x0, x1, measured_data(10000)

      num_variables = 3
      length_work_array = 60025

      algorithm_tolerance = dsqrt( dpmpar(1) )
c
c  Read from standard input the number of data points and the name of
c  the scan data file.
c

      read(5,50) num_points
 50   format(i10)

c     read(5,55) filename
c55   format(1x,a)

      read(5,55) filename
 55   format(a)

c     write(6,1000) num_points, filename
c1000 format( ' num_points = ', i8, '  filename = "', a, '"' )

      if ( num_points .gt. 10000 ) then
            write(6,1001) ' 0 0'
 1001       format(a)
            stop ''
      endif

c
c  Read the data from the data file.
c
      open( unit=1, name=filename, status='OLD', form='FORMATTED',
     +                                      iostat=errflag )

      if ( errflag .ne. 0 ) then
            write(6,1002) ' 0 1 ', errflag
 1002       format(a, i8)
            stop ''
      endif

      state = 0
      min_value =  1.0e38
      max_value = -1.0e38
      min_position = 0.0
      max_position = 0.0

      do 100 i = 1, num_points
            read(1,*,end=9999,err=9999) xvalue(i), measured_data(i)

            if ( state .eq. 0 ) then
                  xstart = xvalue(i)
                  state = 1
            else if ( state .eq. 1 ) then
                  xnext = xvalue(i)
                  state = 2
            endif

            if ( measured_data(i) .lt. min_value ) then
                  min_value = measured_data(i)
                  min_position = i
            endif
            if ( measured_data(i) .gt. max_value ) then
                  max_value = measured_data(i)
                  max_position = i
            endif
 100  continue

      close(1)
c
c  Find the approximate location of the half maximum point.
c
      half_max_value = min_value + 0.5 * ( max_value - min_value )

      closest_to_half_max = 0
      second_closest_to_half_max = 0
      smallest_distance = 9.0e37
      second_smallest_distance = 1.0e38

      do 105 i = 1, num_points
            distance = abs( measured_data(i) - half_max_value )

c           print *, 'i = ', i, '   distance = ', distance

            if ( distance .lt. smallest_distance ) then

                  second_smallest_distance = smallest_distance
                  second_closest_to_half_max = closest_to_half_max

                  smallest_distance = distance
                  closest_to_half_max = i

c                 print *, 'smallest_distance = ', smallest_distance

c                 print *,
c    +     'closest_to_half_max = ', closest_to_half_max,
c    +   '  second_closest_to_half_max = ', second_closest_to_half_max

            elseif ( distance .lt. second_smallest_distance ) then

                 second_smallest_distance = distance
                 second_closest_to_half_max = i

c                print *, 'second_smallest_distance = ',
c    +                       second_smallest_distance

c                 print *,
c    +     'closest_to_half_max = ', closest_to_half_max,
c    +   '  second_closest_to_half_max = ', second_closest_to_half_max

            endif
 105  continue

      if ( closest_to_half_max .eq. 0 ) then
            write(6,1001) ' 0 0'
            stop ''
      endif

      if ( second_closest_to_half_max .eq. 0 ) then
            half_max_position = closest_to_half_max

c           print *, 'case 1'

      elseif (closest_to_half_max .eq. second_closest_to_half_max) then
            half_max_position = closest_to_half_max

c           print *, 'case 2'

      else
            xa = closest_to_half_max
            ya = measured_data( closest_to_half_max )

            xb = second_closest_to_half_max
            yb = measured_data( second_closest_to_half_max )

c           print *, 'xa = ', xa, '  ya = ', ya
c           print *, 'xb = ', xb, '  yb = ', yb

            slope = ( yb - ya ) / ( xb - xa )

c           print *, 'slope = ', slope

            half_max_position = xa + ( half_max_value - ya ) / slope
      endif

c     half_max_position = 0.5 * float( num_points - 1 ) + 1.0
c
c  Record the coefficients of the equation that relates the array index
c  to the x position:
c
c  x = x0 + x1 * (i - 1)
c
      x0 = xstart
      x1 = xnext - xstart

c     print *, 'x0 = ', x0, '   x1 = ', x1
c     print *, 'min_position = ',min_position,'min_value = ',min_value
c     print *, 'max_position = ',max_position,'max_value = ',max_value
c     print *, 'half_max_position = ',half_max_position
c     print *, 'half_max_value = ',half_max_value
c
c  Initialize fitting parameters.
c
      if ( max_position .lt. min_position ) then
            b(1) = max_value
            b(2) = min_value
            b(3) = half_max_position
      else
            b(1) = min_value
            b(2) = max_value
            b(3) = half_max_position
      endif
c
c  Invoke the Levenberg-Marquardt fitting routine LMDIF1 from the
c  Minpack distribution.
c

      call lmdif1( fit_function, num_points, num_variables,
     +                b, function_array,
     +                algorithm_tolerance, algorithm_info,
     +                int_work_array, work_array,
     +                length_work_array )

      euclidean_norm = enorm( num_points, function_array )

c     print *, 'euclidean_norm = ', euclidean_norm
c     print *, 'algorithm_info = ', algorithm_info
c     print *, ' '
c     print *, ' '

c
c  Convert the fitting parameters back from array index units
c  to engineering units.
c
      a(1) = b(1)
      a(2) = b(2)
      a(3) = x0 + x1 * b(3)
c
c  Write the results of the fit to standard output.
c
      x_simple = x0 + x1 * ( half_max_position - 1.0 )

      write(6,1020) euclidean_norm, algorithm_info, x_simple,
     +                           ( a(i), i = 1, num_variables )
 1020 format( ' 1 ', 1P, D14.6, ' ', I2, ' ', 5D14.6 )

      stop ''

 9999 write(6,9001) ' 0 999'
 9001 format(a)
      stop ''
      end
c
c**********************************************************************
c
      subroutine fit_function( num_points, num_variables,
     +                          b, function_array,
     +                          iflag )
      implicit none

      integer i, j, num_points, num_variables, iflag
      real*8 b( num_variables )
      real*8 function_array( num_points )
      real*8 x0, x1, measured_data

      common /optimize/x0, x1, measured_data(10000)

c     write(6,2000) ( b(i), i = 1, num_variables )
c2000 format( ' ', 5D15.7 )

      do 200 i = 1, num_points

            j = i - 1

            if ( j .lt. b(3) ) then
                  function_array(i) = measured_data(i) - b(1) 
            else
                  function_array(i) = measured_data(i) - b(2) 
            endif

 200  continue

      return
      end

